from dataspree.inference.core.consumer import Consumer
from dataspree.inference.core.decorators import action, setter, serializer
from dataspree.inference.core.plugin import Plugin
from dataspree.inference.core.producer import Producer
from dataspree.inference.core.item import Item

from typing import Callable, Dict, Optional, Tuple, List

import time
import logging

logger = logging.getLogger(__name__)


class OrchestratorPlugin(Plugin):

    def __init__(self, name: str, sort_by_time_strictly: bool = False) -> None:
        super(OrchestratorPlugin, self).__init__(name)

        self.flush: bool = False
        self.flush_on_full: bool = False
        self.__consumers: List[Tuple[Consumer, Callable[[Item], Optional[Item]], bool, Optional[Producer]]] = []

        self.__sort_by_time_strictly: bool = sort_by_time_strictly
        self.__last_timestamp: float = 0.0

        self.__events: Dict[int, List[Callable[[int], None]]] = {}

    @setter
    def set_sort_by_time_strictly(self, sort_by_time_strictly: bool) -> None:
        self.__last_timestamp = 0.0
        self.__sort_by_time_strictly = sort_by_time_strictly

    @serializer('sort_by_time_strictly')
    def set_sort_by_time_strictly(self) -> bool:
        return self.__sort_by_time_strictly

    def register_pipeline(self, consumer: Consumer, callback: Callable[[Item], Optional[Item]], *,
                          required_strict: bool = True, producer: Optional[Producer] = None) -> None:
        self.__consumers.append((consumer, callback, required_strict, producer))
        self.add_consumer(consumer)
        if producer is not None and producer.name not in self.producers:
            self.add_producer(producer)

    def schedule_event_after(self, timestamp: int, method: Callable[[int], None]):
        """
        Schedule execution of an event before the next item with a timestamp >= the provided timestamp is executed.

        Must not be used if:
        1. The event should strictly be executed around the given timestamp (execution is event-based, and only if
            a new event is ready to be executed with a timestamp later than the given timestamp)
        2. Execution of :method: has an effect on :self.__consumers:. This would create OOO samples e.g., in
           synchronous mode or can cause a deadlock:

           C1:   10
           C2:   20
           C3:   30
           E:    after t=5: do stuff and send it out to C3 with the timestamp

           E would be executed prior to C1, and would append an event to C3.  If the queue size is 1, a deadlock occurs.
           If the queue size is larger, the queues look as follows:

           C1:   10
           C2:   20
           C3:   30, 10(result(e))
           E:    /

           So the execution order would be as follows if nothing is added to the queues:
           C1, C2, C3, C3

           which would be OOO.
        """
        self.__events.setdefault(timestamp, []).append(method)

    def before_item_processed(self, item: Item, consumer: Consumer) -> None:
        """
        Callback that is executed before an item is processed.

        :param item:            The item that is about to be processed as a part of an orchestrator pipeline.
        :param consumer:        The consumer that just emitted the item.
        """
        ...

    def on_item_processed(self, item: Item, consumer: Consumer,
                          processed_item: Optional[Item], producer: Optional[Producer]) -> None:
        ...

    @action
    def trigger_flush(self) -> None:
        self.flush = True

    def run(self) -> None:
        for item, consumer, callback, producer in self.__get_next_item():
            current_timestamp = item['timestamp']

            # check if an event has to be processed:
            if len(self.__events):
                sorted_events = sorted(self.__events.items())
                timestamp_event = sorted_events[0][0]

                if timestamp_event < current_timestamp:
                    self.__last_timestamp = max(self.__last_timestamp or timestamp_event, timestamp_event)

            # Completely in sync in full replay mode. In on-site-deployment slight OOO occurs because we have
            # multiple consumers.
            if __debug__:
                if self.__sort_by_time_strictly and not self.flush and not self.flush_on_full and \
                        not any(True for r in self.__consumers if not r[2]):
                    assert current_timestamp >= self.__last_timestamp
            self.__last_timestamp = current_timestamp

            self.before_item_processed(item, consumer)
            processed_item: Optional[Item] = callback(item)
            if producer is not None and processed_item is not None:
                producer.send(processed_item)

            self.on_item_processed(item, consumer, processed_item, producer)

        logger.info(f'Orchestrator Plugin: run: Shutting down orchestrator: running={self.running}.')
        assert not self.running

    def __get_next_item(self) -> Optional[Tuple[Item, Consumer]]:
        new_items: Dict[str, Item] = dict()
        while self.running:

            #
            # Fill next value(s) into new_items
            #

            if self.__sort_by_time_strictly:

                # Replay setup: Wait for at least one item in each queue, unless flush or that consumer is
                # "not required".
                remaining_consumers: List[Tuple[Consumer, bool]] = \
                    [(consumer, required) for (consumer, _, required, _) in self.__consumers
                     if consumer.name not in new_items]

                while len(remaining_consumers):

                    next_remaining_consumers: List[Tuple[Consumer, bool]] = []
                    any_full: bool = any(consumer.is_full() for (consumer, _, _, _) in self.__consumers)

                    for (consumer, required) in remaining_consumers:
                        potential_next_item: Item = consumer.receive(block=False, idle_sleep_override=0)

                        if potential_next_item is None:
                            # Still remaining
                            if self.flush:
                                continue

                            if any_full and self.flush_on_full:
                                continue

                            if not required:
                                continue

                            next_remaining_consumers.append((consumer, required))

                        else:
                            # Not remaining anymore, add to items
                            assert consumer.name not in new_items
                            new_items[consumer.name] = potential_next_item

                    remaining_consumers = next_remaining_consumers

                    if not self.running:
                        return None

                    if len(remaining_consumers):
                        time.sleep(0.05)

                assert not len(remaining_consumers)

            else:

                # Productive setup: Fill in all values that are available yet, and use their value
                for (consumer, _, _, _) in self.__consumers:
                    if consumer.name not in new_items:
                        item: Item = consumer.receive(block=False, idle_sleep_override=0)
                        if item is not None:
                            new_items[consumer.name] = item

                time.sleep(0.005)

            #
            # Process first new item
            #

            if len(new_items):
                # noinspection PyTypeChecker
                sorted_items: List[Tuple[str, Item]] = sorted(new_items.items(), key=lambda a: a[1]['timestamp'])
                item = sorted_items[0][1]

                consumers = [c for c in self.__consumers if c[0].name == sorted_items[0][0]]
                assert len(consumers) == 1

                (consumer, callback, _, producer) = consumers[0]
                del new_items[sorted_items[0][0]]
                yield item, consumer, callback, producer

            else:
                time.sleep(0.05)

    @classmethod
    def get_config_entries(cls) -> Dict:
        return {
            'fields': [
                {
                    'name': 'sort_by_time_strictly',
                    'display_name': 'Strictly chronological',
                    'description': 'Set this option to wait for an input for all required consumers until the next item'
                                   'is forwarded to any processing function / producer. '
                                   'This is only feasible in replay mode, or if it guaranteed that all inputs receive '
                                   'data regularly and at a high enough frequency.'
                }
            ]
        }

    @staticmethod
    def get_config_categories() -> List[str]:
        return ['plugins']
