from typing import Optional, Dict

import cv2
import copy
import numba
import numpy as np

from dataspree.inference.core.config import ConfigFieldTypes
from dataspree.inference.core.decorators import action
from dataspree.inference.core.item import Item
from dataspree.inference.core.localization_utils import ObjectPosition, AlignedPointCloud
from dataspree.inference.core.producer import Property, PropertyType, Producer
from dataspree.inference.core.simple_plugin import SimplePlugin

from numba.pycc import CC


class CalibrationVisualization(SimplePlugin):

    def __init__(self, name: str, pallet: ObjectPosition, tolerance: int = 30,
                 apply_calibration: bool = False, filter_for_largest_component: bool = True) -> None:

        SimplePlugin.__init__(self, name, [Property('image', PropertyType.IMAGE)])

        self.pallet: ObjectPosition = pallet
        self.tolerance: int = tolerance

        self.apply_calibration: bool = apply_calibration

        self.filter_for_largest_component: bool = filter_for_largest_component

        self.producer_pallet: Producer = Producer(f'{self.name}_pallet', [Property('image', PropertyType.IMAGE)])
        self.add_producer(self.producer_pallet)

        # log scale
        self.producer_pallet_log: Producer = Producer(f'{self.name}_pallet_log',
                                                      [Property('image', PropertyType.IMAGE)])
        self.add_producer(self.producer_pallet_log)

        self.__item = None

    def process_item(self, item: Item) -> Optional[Item]:
        if 'point_cloud' not in item:
            return None

        if self.__item == None:
            self.__item = item

        # item = copy.deepcopy(self.__item)

        # rotation: ObjectRotation = item.get('rotation', self.rotation)
        # pallet_center = item.get('origin', np.array([self.offset_x, self.offset_y, self.offset_z]).astype(np.int16))

        pallet_width: int = 1200
        pallet_height: int = 800

        point_cloud = item['point_cloud'].copy()

        # Apply the rotation & translation
        if self.apply_calibration:
            aligned: AlignedPointCloud = self.pallet.align(point_cloud)
            point_cloud = item['point_cloud'] = aligned.aligned_cloud
            item['invalid_height'] = aligned.invalid_height

        else:
            if 'invalid_height' not in item:
                item['invalid_height'] = np.zeros(point_cloud.shape[:-1], dtype=np.uint8)

        display_image = np.zeros((*point_cloud.shape[:-1], 3), np.uint8)
        min_amp = item['amplitude_image'].min()
        max_amp = item['amplitude_image'].max()
        display_image[:, :, 1] = (item['amplitude_image'].astype(float, copy=True) - min_amp) / \
                                 (max_amp - min_amp) * 255

        display_image, distance_image, binary_image = get_distance_binary_display_image(display_image, point_cloud,
                                                                                        item['invalid_height'],
                                                                                        pallet_width, pallet_height,
                                                                                        self.tolerance)

        # Exclude everything that is not connected to the largest component (supposedly the pallet)
        if self.filter_for_largest_component:
            nb_components, output, stats, _centroids = cv2.connectedComponentsWithStats(binary_image, connectivity=4)
            if nb_components:
                max_label, max_size = max(((i, stats[i, cv2.CC_STAT_AREA]) for i in range(1, nb_components)),
                                          key=lambda x: x[1])
                binary_image = output == max_label

        # scale the distance channel in the normal display image
        max_dist = np.max(distance_image)
        min_dist = np.min(distance_image)
        distance_image_uint8 = np.where(
            binary_image, ((distance_image - min_dist).astype(float) * 255 / (max_dist - min_dist)).astype(np.uint8), 0)
        # display_image[..., 0] = distance_image_uint8

        min_ix, max_ix, min_iy, max_iy = get_minmax(binary_image)
        pallet_image = np.where(binary_image[min_ix:max_ix, min_iy:max_iy],
                                distance_image[min_ix:max_ix, min_iy:max_iy], 0)

        # pallet image: heatmap
        abs_pallet_image = np.abs(pallet_image.astype(np.int64)).astype(np.uint64)
        exp_pallet_image = np.exp(abs_pallet_image)

        def color_map(img):
            mn, mx = np.min(img), np.max(img)
            mn = -np.max(-np.where(img > 1, img, np.infty))
            margin = 50
            one_channel_img = np.zeros((img.shape[0], img.shape[1] + margin), np.uint8)
            one_channel_img[:, :-margin] = np.where(img != 0,
                                                    ((img - mn).astype(float) / (mx - mn) * 255).astype(np.uint8), 0)
            one_channel_img[:, -margin:] = np.linspace(0, 255, img.shape[0])[..., None]

            color_mapped = cv2.applyColorMap(one_channel_img, cv2.COLORMAP_JET)
            color_mapped[:, :-margin] = np.where((img != 0)[..., None], color_mapped[:, :-margin], (0, 0, 0))
            cv2.putText(color_mapped, f'{mn}', (img.shape[1], 25),
                        cv2.FONT_HERSHEY_TRIPLEX, 0.45, (255, 255, 255), 1, cv2.LINE_AA)
            cv2.putText(color_mapped, f'{mx}', (img.shape[1], img.shape[0] - 25),
                        cv2.FONT_HERSHEY_TRIPLEX, 0.45, (255, 255, 255), 1, cv2.LINE_AA)
            return color_mapped

        if abs_pallet_image.shape[0] and abs_pallet_image.shape[1]:
            display_image_abs_pallet = color_map(abs_pallet_image)
            display_image_exp_pallet = color_map(exp_pallet_image)

            self.producer_pallet.send(Item.default(display_image=display_image_abs_pallet))
            self.producer_pallet_log.send(Item.default(display_image=display_image_exp_pallet))

        item['display_image'] = display_image

        # self.producers['calibr'].send(item)

        self.producers[self.name].send(item)
        print(item['source_id'])

        return None

    @action
    def calibrate(self) -> None:
        # TODO Auto calibrate
        raise NotImplementedError('not implemented.')

    @classmethod
    def get_config_entries(cls) -> Dict:
        return {
            'display_name': 'Safety controller',
            'description': 'Plugin for running the safety module only',
            'updatable': True,
            'fields': [
                ObjectPosition.get_config_entries('pallet'),
                {
                    'name': 'tolerance',
                    'display_name': 'Tolerance [mm]',
                    'type': ConfigFieldTypes.INTEGER_INPUT,
                    'default': 30,
                    'description': ''
                },
                {
                    'name': 'apply_calibration',
                    'display_name': 'Apply calibration',
                    'type': ConfigFieldTypes.BOOL_INPUT,
                    'default': False,
                    'description': ''
                },
                {
                    'name': 'filter_for_largest_component',
                    'display_name': 'Filter for largest component',
                    'type': ConfigFieldTypes.BOOL_INPUT,
                    'default': True,
                    'description': ''
                }
            ]
        }

    @action
    def roll_p(self):
        self.pallet.rotation.roll += 0.01

    @action
    def roll_pp(self):
        self.pallet.rotation.roll += 0.01

    @action
    def pitch_p(self):
        self.pallet.rotation.pitch += 0.1

    @action
    def pitch_pp(self):
        self.pallet.rotation.pitch += 0.1

    @action
    def yaw_p(self):
        self.pallet.rotation.yaw += 0.01

    @action
    def yaw_pp(self):
        self.pallet.rotation.yaw += 0.1

    @action
    def roll_m(self):
        self.pallet.rotation.roll -= 0.01

    @action
    def roll_mm(self):
        self.pallet.rotation.roll -= 0.1

    @action
    def pitch_m(self):
        self.pallet.rotation.pitch -= 0.01

    @action
    def pitch_mm(self):
        self.pallet.rotation.pitch -= 0.1

    @action
    def yaw_m(self):
        self.pallet.rotation.yaw -= 0.01

    @action
    def yaw_mm(self):
        self.pallet.rotation.yaw -= 0.1


cc = CC('calibration_visu')


@numba.njit
@cc.export('get_distance_binary_display_image',
           numba.types.Tuple((numba.uint8[:, :, :], numba.int16[:, :], numba.uint8[:, :]))(
               numba.uint8[:, :, :], numba.int16[:, :, :], numba.uint8[:, :], numba.int16, numba.int16, numba.int16))
def get_distance_binary_display_image(display_image: np.ndarray, point_cloud: np.ndarray, invalid: np.ndarray,
                                      pallet_width: int, pallet_height: int, tolerance: int):
    binary_image = np.zeros_like(display_image[..., 0])
    distance_image = np.zeros(point_cloud.shape[:-1], np.int64)

    for i_x, pcs in enumerate(point_cloud):
        for i_y, pc in enumerate(pcs):
            assert pc.shape == (3,)
            x, y, z = pc

            if invalid[i_x, i_y]:
                continue

            if (0 <= np.abs(x) <= pallet_width / 2) and \
                    (0 <= np.abs(y) <= pallet_height / 2):

                if 0 <= np.abs(x) <= pallet_width / 2 - tolerance and \
                        0 <= np.abs(y) <= pallet_height / 2 - tolerance:

                    distance_image[i_x, i_y] = z
                    binary_image[i_x, i_y] = 1
                    display_image[i_x, i_y, 0] = 0

                else:
                    display_image[i_x, i_y, 0] = 125

            else:
                display_image[i_x, i_y, 0] = 255

    return display_image, distance_image, binary_image


@numba.njit
@cc.export('get_minmax',
           numba.types.Tuple((numba.uint16, numba.uint16, numba.uint16))(numba.uint8[:, :]))
def get_minmax(binary_image):
    min_ix, max_ix = binary_image.shape[0], 0
    min_iy, max_iy = binary_image.shape[1], 0
    for i_x, xs in enumerate(binary_image):
        for i_y, p in enumerate(xs):
            if p:
                min_ix = min(min_ix, i_x)
                max_ix = max(max_ix, i_x)
                min_iy = min(min_iy, i_y)
                max_iy = max(max_iy, i_y)

    return min_ix, max_ix, min_iy, max_iy
