import cv2
import numpy as np
from scipy.stats import chi2
from typing import Tuple, List, Optional, Union

from dataspree.inference.core.environment import is_application_debug

__LOG_2PI: float = np.log(2 * np.pi)

def get_covariance_confidence_ellipse(covariance_matrix: np.ndarray,
                                      probabilities: List[float]) -> Tuple[np.ndarray, float]:
    state_dimensionality = covariance_matrix.shape[0]
    assert covariance_matrix.shape == (state_dimensionality, state_dimensionality)
    assert state_dimensionality == 2, 'This should be a requirement for acquiring the rotation.'

    number_of_ellipses = len(probabilities)

    # Once rotated such that the covariance is zero, the distribution of the RV's components (x, y, (z...) coordinate)
    # is independent of the other components and distributed according to a normal distribution
    # N( (.)  ; \\mu, eigen_value).

    # The sum of squared Gaussians x \\in \\R, y \\in \\R, ... is Chi Square distributed.
    values = np.array([np.sqrt(chi2.ppf(probability, state_dimensionality)) for probability in probabilities])
    assert values.shape == (number_of_ellipses,)

    eigen_values, eigen_vectors = np.linalg.eig(covariance_matrix)
    assert eigen_values.shape == (state_dimensionality,)
    assert eigen_vectors.shape == (state_dimensionality, state_dimensionality)

    rotation = -np.arctan2(*eigen_vectors[0][::-1])

    sizes = values[:, np.newaxis] * np.sqrt(eigen_values[np.newaxis]) * 2
    assert sizes.shape == (number_of_ellipses, state_dimensionality)
    return sizes, rotation


def draw_covariance_ellises(image, means, covariance_matrices, probabilities: List[float], colors: List, rs,
                            image_size: Optional[Tuple[int, int]] = None,
                            original_image_size: Optional[Tuple[int, int]] = None):
    scale_factor: np.ndarray = np.ones(2) if original_image_size is None or image_size is None \
        else np.array(image.shape[:2], dtype=np.float32) / np.array(original_image_size, dtype=np.float32)

    number_of_distributions: int = means.shape[0]
    state_dimensionality: int = means.shape[1]

    assert means.shape == (number_of_distributions, state_dimensionality)
    assert covariance_matrices.shape == (number_of_distributions, state_dimensionality, state_dimensionality)
    assert len(colors) == number_of_distributions
    assert len(probabilities) > 0

    np.multiply(means[:, :2], scale_factor, out=means[:, :2])
    for i_d, mean in enumerate(means):

        try:
            sizes, rotation = get_covariance_confidence_ellipse(covariance_matrices[i_d][:2, :2], probabilities)
            sizes = sizes * scale_factor

            for i_p, size in enumerate(sizes):

                center = tuple(np.round(mean[:2]).astype(int))
                image = cv2.ellipse(image, center, tuple(np.round(size).astype(int)),
                                    (rotation + np.pi) / np.pi * 90, 0, 360, colors[i_d], 1)

                if rs is not None:
                    image = cv2.putText(image, f'{rs[i_d]:.3f}', center, cv2.FONT_HERSHEY_SIMPLEX,
                                        0.5 * scale_factor.min(initial=1), thickness=1, color=(0, 0, 0))
        except Exception as e:
            print(e)


def inplace_mod(values: np.ndarray, mod_value: float) -> None:
    np.mod(values, mod_value, out=values)

    np.subtract(values, mod_value / 1, out=values, where=values >= mod_value / 2)

    """
    if not (values <= mod_value / 2).all():
        print('hier')
    assert (values <= mod_value / 2).all()
    assert (values >= -mod_value / 2).all()
    """
    return values


def evaluate_gaussians(x: np.ndarray, means: np.ndarray, covariance_matrices: np.ndarray,
                       inverted_covariance_matrices: Optional[np.ndarray] = None,
                       divisors=None, return_components: bool = False,
                       dimensions_modulo: Optional[List[Tuple[int, float]]] = None) \
        -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:
    number_of_values: int = x.shape[0]
    x_dimensionality: int = x.shape[1]
    assert x.shape == (number_of_values, x_dimensionality)

    number_of_distributions: int = means.shape[0]
    state_dimensionality: int = means.shape[1]
    assert means.shape == (number_of_distributions, state_dimensionality)
    assert covariance_matrices.shape == (number_of_distributions, state_dimensionality, state_dimensionality)

    if inverted_covariance_matrices is None:
        inverted_covariance_matrices = np.linalg.inv(covariance_matrices)
    assert inverted_covariance_matrices.shape == (number_of_distributions, state_dimensionality, state_dimensionality)

    if divisors is None:
        divisors = np.sqrt((2 * np.pi) ** state_dimensionality * np.linalg.det(covariance_matrices))
    assert divisors.shape == (number_of_distributions,)

    difference_from_mean: np.ndarray = np.zeros((number_of_distributions, number_of_values, state_dimensionality),
                                                dtype=means.dtype)
    difference_from_mean[:, :, :x_dimensionality] = x[None] - means[:, :x_dimensionality][:, None]
    assert difference_from_mean.shape == (number_of_distributions, number_of_values, state_dimensionality)

    # Make "mod" if necessary (e.g. for angle)
    if dimensions_modulo is not None:
        for mod_index, mod_value in dimensions_modulo:
            test = inplace_mod(difference_from_mean[:, :, mod_index], mod_value)
            assert (test == difference_from_mean[:, :, mod_index]).all()

    tmp = np.einsum('dvS,dsS,dvs->dv', difference_from_mean, inverted_covariance_matrices, difference_from_mean)
    assert tmp.shape == (number_of_distributions, number_of_values)

    values: np.ndarray = np.exp(- 0.5 * tmp) / divisors[:, None]
    assert values.shape == (number_of_distributions, number_of_values)

    if __debug__ and is_application_debug():
        __debug_values: np.ndarray = np.array([[
            np.exp(-.5 * (
                np.pad(x[i_v] - means[i_d][:x_dimensionality], (0, state_dimensionality - x_dimensionality))).T.dot(
                inverted_covariance_matrices[i_d]).dot(
                np.pad(x[i_v] - means[i_d][:x_dimensionality], (0, state_dimensionality - x_dimensionality)))) /
            np.sqrt((2 * np.pi) ** state_dimensionality * np.linalg.det(covariance_matrices[i_d]))
            for i_v in range(number_of_values)] for i_d in range(number_of_distributions)]).reshape((
            number_of_distributions, number_of_values
        ))

        __diff = np.abs(__debug_values - values)
        # assert (__diff >= 0).all() and (np.max(__diff, initial=0) < 1e-9) XXX

    if return_components:
        return values, inverted_covariance_matrices, divisors, difference_from_mean
    return values

# This could be jitted.
def multi_psd(covariance_matrices: np.ndarray, allow_singular: bool = True):
    number_of_pdfs: int = covariance_matrices.shape[0]
    state_dimensionality: int = covariance_matrices.shape[1]
    assert covariance_matrices.shape == (number_of_pdfs, state_dimensionality, state_dimensionality)

    # eigen decomposition of the covariance matrices.
    # The eigen_vector matrix contains an eigenvector matrix for each pdf, having the eigen_vector in the rows
    # [
    #  [
    #   eigen_vector_0 #<  row 0 (state_dimensionality)
    #   eigen_vector_1 #<  row 1
    #  ]
    # ]
    eigen_values, eigen_vectors = np.linalg.eigh(covariance_matrices)
    assert eigen_values.shape == (number_of_pdfs, state_dimensionality)
    assert eigen_vectors.shape == (number_of_pdfs, state_dimensionality, state_dimensionality)

    if __debug__ and is_application_debug():
        #
        av = np.einsum('NsS,NSd->Nsd', covariance_matrices, eigen_vectors)

        # eigen_values[pdf id, [[broadcast]], state_dimensionality) * eigen_vectors
        #
        lambda_v = eigen_values[:, np.newaxis] * eigen_vectors
        __diff = av - lambda_v
        assert np.all(np.abs(__diff) < 1e-9)

    # Find out small value
    s_dtype: str = eigen_values.dtype.char.lower()
    eps_per_pdf: np.ndarray = np.max(np.abs(eigen_values), axis=1) \
                              * ((1E3 if s_dtype == 'f' else 1E6) * np.finfo(s_dtype).eps)
    assert eps_per_pdf.shape == (number_of_pdfs,)

    min_ev_per_pdf: np.ndarray = np.min(eigen_values, axis=1)
    assert min_ev_per_pdf.shape == (number_of_pdfs,)

    if np.any(min_ev_per_pdf < -eps_per_pdf):
        raise ValueError('The covariance matrix must be positive semidefinite. It is not.')

    where_positive: np.ndarray = eigen_values > eps_per_pdf[:, np.newaxis]
    log_pseudo_determinant: np.ndarray = np.sum(np.log(eigen_values), where=where_positive, initial=0, axis=1)
    ranks: np.ndarray = np.sum(where_positive, axis=1, initial=0)
    if not allow_singular and np.any(ranks < state_dimensionality):
        raise ValueError('singular covariance matrix.')

    # eigen_values \gets 1 / eigen_values
    where_ind = np.abs(eigen_values) > eps_per_pdf[:, np.newaxis]
    eigen_values[~where_ind] = 0
    np.reciprocal(eigen_values, where=where_ind, out=eigen_values)
    np.sqrt(eigen_values, where=where_ind, out=eigen_values)

    # eigen_vectors \gets eigen_vectors / eigen_value-orig
    # -> Mat * eigen_vector_new = Mat * eigen_vector-orig / eigen_value-orig
    #                           = eigen_value-orig * eigen_vector-orig / eigen_value-orig
    #                           = eigen_vector-orig
    # Nve x Nl:
    np.multiply(eigen_vectors, eigen_values[:, np.newaxis], out=eigen_vectors)

    # ev_new \\cdot ev_new.T
    # This is the inverse (M orig cov, V = ev evT:
    # M V V.T = (Evs...) \\cdot V.T = I ( * 1 / ev?)
    pseudo_inverse: np.ndarray = np.einsum('nsS,naS->nsa', eigen_vectors, eigen_vectors)

    if __debug__ and is_application_debug():
        for i, covariance_matrix in enumerate(covariance_matrices):
            __I1 = covariance_matrix.dot(pseudo_inverse[i])
            __I2 = pseudo_inverse[i].dot(covariance_matrix)

            __diff = __I1 - np.eye(__I1.shape[0])
            assert np.all(np.abs(__diff) < 1e-9)

            __diff = __I2 - np.eye(__I2.shape[0])
            assert np.all(np.abs(__diff) < 1e-9)

    return ranks, eigen_vectors, log_pseudo_determinant, pseudo_inverse


def log_evaluate_gaussians_2(x: np.ndarray, means: np.ndarray, covariance_matrices: np.ndarray,
                           dimensions_modulo: Optional[List[Tuple[int, float]]] = None):

    number_of_values: int = x.shape[0]
    x_dimensionality: int = x.shape[1]
    assert x.shape == (number_of_values, x_dimensionality)

    number_of_distributions: int = means.shape[0]
    state_dimensionality: int = means.shape[1]
    assert means.shape == (number_of_distributions, state_dimensionality)
    assert covariance_matrices.shape == (number_of_distributions, state_dimensionality, state_dimensionality)

    ranks, mat_u, log_pseudo_determinant, pseudo_inverse = multi_psd(covariance_matrices)
    assert ranks.shape == (number_of_distributions,)
    assert mat_u.shape == (number_of_distributions, state_dimensionality, state_dimensionality)
    assert log_pseudo_determinant.shape == (number_of_distributions,)

    # Quasi outer sum of x and means, such that the subtraction is performed along the
    # dimensionality and broadcasted along both other axes.
    difference_from_mean: np.ndarray = np.zeros((number_of_distributions, number_of_values, state_dimensionality),
                                                dtype=means.dtype)
    difference_from_mean[:, :, :x_dimensionality] = x[None] - means[:, :x_dimensionality][:, None]
    assert difference_from_mean.shape == (number_of_distributions, number_of_values, state_dimensionality)

    # Make "mod" if necessary (e.g. for angle)
    if dimensions_modulo is not None:
        for mod_index, mod_value in dimensions_modulo:
            test = inplace_mod(difference_from_mean[:, :, mod_index], mod_value)
            assert (test == difference_from_mean[:, :, mod_index]).all()

    maha = np.einsum('DVd,Dds->DVs', difference_from_mean, mat_u)
    maha = -0.5 * (np.sum(maha ** 2, axis=2) + ranks[:, np.newaxis] * __LOG_2PI
                   + log_pseudo_determinant[:, np.newaxis])
    # Alternative if operators can be fused
    # np.square(diff, out=diff)
    # maha = np.sum(diff, axis=2)
    # np.add(maha, ranks[:, np.newaxis] * __LOG_2PI + log_pseudo_determinant[:, np.newaxis], out=maha)
    # np.multiply(maha, -0.5, out=maha)

    assert maha.shape == (number_of_distributions, number_of_values)

    return maha, pseudo_inverse, difference_from_mean


def evaluate_gaussians(x: np.ndarray, means: np.ndarray, covariance_matrices: np.ndarray):
    return np.exp(log_evaluate_gaussians(x, means, covariance_matrices))



def log_evaluate_gaussians(x: np.ndarray, means: np.ndarray, covariance_matrices: np.ndarray,
                           inverted_covariance_matrices: Optional[np.ndarray] = None,
                           subtractors=None, return_components: bool = False,
                           dimensions_modulo: Optional[List[Tuple[int, float]]] = None) \
        -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]]:
    number_of_values: int = x.shape[0]
    x_dimensionality: int = x.shape[1]
    assert x.shape == (number_of_values, x_dimensionality)

    number_of_distributions: int = means.shape[0]
    state_dimensionality: int = means.shape[1]
    assert means.shape == (number_of_distributions, state_dimensionality)
    assert covariance_matrices.shape == (number_of_distributions, state_dimensionality, state_dimensionality)

    if inverted_covariance_matrices is None:
        inverted_covariance_matrices = np.linalg.inv(covariance_matrices)
    assert inverted_covariance_matrices.shape == (number_of_distributions, state_dimensionality, state_dimensionality)

    if subtractors is None:
        # log( sqrt((2pi)**s *d))  = log ( ((2pi)**s * d)**.5) ) = 0.5 [log((2 pi) **s) + log(d)] =
        #  0.5 ( s * log(2pi) + log(d)
        subtractors = 0.5 * np.log(np.linalg.det((2 * np.pi) * covariance_matrices))
    assert subtractors.shape == (number_of_distributions,)

    # Quasi outer sum of x and means, such that the subtraction is performed along the
    # dimensionality and broadcasted along both other axes.
    difference_from_mean: np.ndarray = np.zeros((number_of_distributions, number_of_values, state_dimensionality),
                                                dtype=means.dtype)
    difference_from_mean[:, :, :x_dimensionality] = x[None] - means[:, :x_dimensionality][:, None]
    assert difference_from_mean.shape == (number_of_distributions, number_of_values, state_dimensionality)

    # Make "mod" if necessary (e.g. for angle)
    if dimensions_modulo is not None:
        for mod_index, mod_value in dimensions_modulo:
            test = inplace_mod(difference_from_mean[:, :, mod_index], mod_value)
            assert (test == difference_from_mean[:, :, mod_index]).all()

    # difference^T \\cdot inverted_covariance_matrices \\cdot difference
    tmp = np.einsum('dvS,dsS,dvs->dv', difference_from_mean, inverted_covariance_matrices, difference_from_mean)
    assert tmp.shape == (number_of_distributions, number_of_values)

    values = - 0.5 * tmp - subtractors[:, None]
    assert values.shape == (number_of_distributions, number_of_values)

    if __debug__ and is_application_debug():
        __debug_values: np.ndarray = np.empty_like(values)
        for i_d in range(number_of_distributions):
            for i_v in range(number_of_values):
                diff = np.pad(x[i_v] - means[i_d][:x_dimensionality], (0, state_dimensionality - x_dimensionality))

                if dimensions_modulo is not None:
                    for mod_index, mod_value in dimensions_modulo:
                        if diff[mod_index] > mod_value / 2:
                            diff[mod_index] = mod_value - diff[mod_index]

                __debug_values[i_d, i_v] = -0.5 * diff.T.dot(inverted_covariance_matrices[i_d]).dot(diff) \
                                           - np.log(np.sqrt((2 * np.pi) ** state_dimensionality
                                                            * np.linalg.det(covariance_matrices[i_d])))

        __diff = np.abs(__debug_values - values)
        #assert (__diff >= 0).all() and (np.max(__diff, initial=0) < 1e-9) #XXX

    if return_components:
        return values, inverted_covariance_matrices, subtractors, difference_from_mean
    return values


def evaluate_gaussian(x: np.ndarray, mean: np.ndarray, covariance_matrix: np.ndarray,
                      inverted_covariance_matrix: Optional[np.ndarray] = None,
                      divisor: Optional[float] = None, return_components: bool = False,
                      dimensions_modulo: Optional[List[Tuple[int, float]]] = None) \
        -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, float, np.ndarray]]:
    inverted_covariance_matrix = None if inverted_covariance_matrix is None else inverted_covariance_matrix[None]
    divisor = None if divisor is None else np.array(divisor)

    result = evaluate_gaussians(x=x, means=mean[np.newaxis], covariance_matrices=covariance_matrix[np.newaxis],
                                inverted_covariance_matrices=inverted_covariance_matrix,
                                divisors=divisor, return_components=return_components)
    if return_components:
        return result[0], result[1], float(result[2]), result[3]
    return result[0]


def log_evaluate_gaussian(x: np.ndarray, mean: np.ndarray, covariance_matrix: np.ndarray,
                          inverted_covariance_matrix: Optional[np.ndarray] = None,
                          subtractor: Optional[float] = None,
                          return_components: bool = False,
                          dimensions_modulo: Optional[List[Tuple[int, float]]] = None) \
        -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray, float, np.ndarray]]:
    inverted_covariance_matrix = None if inverted_covariance_matrix is None else inverted_covariance_matrix[None]
    subtractor = None if subtractor is None else np.array(subtractor)

    result = log_evaluate_gaussians(x=x, means=mean[np.newaxis], covariance_matrices=covariance_matrix[np.newaxis],
                                    inverted_covariance_matrices=inverted_covariance_matrix, subtractors=subtractor,
                                    return_components=return_components, dimensions_modulo=dimensions_modulo)
    if return_components:
        return result[0], result[1], float(result[2]), result[3]
    return result[0]
