import copy
from functools import partial
from typing import Optional, List, Dict
import time
import cv2
import logging

import numpy as np
from matplotlib import pyplot as plt
from matplotlib import patches as ptch

from dataspree.inference.core.environment import is_application_debug
from dataspree.inference.core.item import Item
from dataspree.inference.core.simple_plugin import SimplePlugin
from dataspree.inference.core.producer import Property, PropertyType
from dataspree.inference.plugins.tracking.mbm import MultibernoulliMixture, GaussianPredictionParametrization, \
    MultibernoulliBirthModel, GaussianTransitionDensity, UpdateModel, get_track_color, SmoothTransitionSurvival, \
    SurvivalProbability, LinearMotionModel
from dataspree.inference.plugins.tracking.gaussian import draw_covariance_ellises

logger = logging.getLogger(__name__)


class TrackingMbm(SimplePlugin):

    def __init__(self, name: str) -> None:
        super(TrackingMbm, self).__init__(name=name, properties=[Property('inference', PropertyType.INFERENCE),
                                                                 Property('image', PropertyType.IMAGE)])

        self.mbm: Optional[MultibernoulliMixture] = None
        self.mbbm: Optional[MultibernoulliBirthModel] = None
        self.t_last: float = 0

        self.image_width = None
        self.image_height = None
        self.last_batch_id: int = 0
        self.dimensionality: int = 0
        self.dimensionality_z: int = 4

        self.draw_confidence: float = 0.9
        self.detection_size_divisor: float = 1.0

        self.last_total_in: int = 0
        self.last_total_out: int = 0

    def instanitate_mbm(self, image_width: int, image_height: int, utilize_velocity: bool = False,
                        utilize_confidence: bool = False):

        self.dimensionality = 2 + (2 if utilize_velocity else 0) + (1 if utilize_confidence else 0)

        fov = (0, 0, image_width, image_height)
        # self.__create_with_size_and_orientation(fov)
        self.__all_with_velocity(fov)
        self.t_last: float = 0

    @staticmethod
    def create_survival_model(p_s, fov, horizontal: bool) -> SurvivalProbability:
        if False:
            retained = 0.7
            retained = 0.99
            return ReducedRoiSurvival(p_s=p_s, fov=fov, horizontal_area_retained=retained if horizontal else 1.0,
                                      vertical_area_retained=retained if not horizontal else 1.0)
        else:
            retained = 0.65
            return SmoothTransitionSurvival(p_s=p_s, fov=fov,
                                            horizontal_area_retained=retained if horizontal else 1.0,
                                            vertical_area_retained=retained if not horizontal else 1.0,
                                            horizontal_percentage_zero=0.2,
                                            vertical_percentage_zero=0.2, min_value=0.2)

    def __create_with_size_and_orientation(self, fov, horizontal: bool = True):
        self.dimensionality_z = 5
        self.dimensionality = 5

        state_pos_columns = (0, 1)
        state_size_columns = (2, 3)
        state_orientation_colum = 4

        self.detection_size_divisor: float = 10.0

        #
        # Prediction:
        #

        # Birth Model
        self.mbbm = MultibernoulliBirthModel.create(fov=fov,
                                                    birth_weight_scale=0.7,
                                                    position_uncertainty=0.10,
                                                    gaussians_per_side=6,
                                                    horizontal=True,
                                                    additional_mean=np.array([5, 5, 0], dtype=np.float),
                                                    additional_cov=np.array(
                                                        [[20, 10, 0], [10, 20, 0], [0, 0, np.pi / 2]]))

        # Survival Probability
        survival_probability: SurvivalProbability = self.create_survival_model(p_s=.9999999, fov=fov,
                                                                               horizontal=horizontal)

        # Transition Density

        scale = 2
        linear_motion_model = LinearMotionModel(np.eye(self.dimensionality))
        noise_covariance = LinearMotionModel(np.diag([3 * scale, 50 * scale, 0.1, 0.1, np.pi / 32]))
        # covariance_transformation_model = LinearMotionModel(np.eye(self.dimensionality))

        gaussian_transition_density = GaussianTransitionDensity(
            linear_motion_model=linear_motion_model, noise_covariance=noise_covariance, rotate_noise_covariance=True,
            state_orientation_column=state_orientation_colum, survival_probability=survival_probability)

        predict = GaussianPredictionParametrization(birth=self.mbbm, transition=gaussian_transition_density)

        #
        # Update
        #
        transform = np.eye(5)
        update_cov = np.diag((30 * scale, 30 * scale, 0.3, 0.3, np.pi / 64))
        update_cov[1, 0] = update_cov[0, 1] = 1 * scale
        update = UpdateModel(transform=transform, covariance=update_cov, p_d=0.3, p_c=1e-11)

        self.mbm = MultibernoulliMixture(predict=predict, update=update, max_number_of_hypothesis=1,
                                         detection_size_divided_by=self.detection_size_divisor)

    def __all_with_velocity(self, fov, horizontal: bool = True):
        self.dimensionality_z = 5
        self.dimensionality = 7

        state_pos_columns = (0, 1)
        state_size_columns = (2, 3)
        state_orientation_colum = 4
        state_velocity_colum = (5, 6)

        self.detection_size_divisor: float = 10.0

        #
        # Prediction:
        #

        # Birth Model
        self.mbbm = MultibernoulliBirthModel.create(fov=fov,
                                                    birth_weight_scale=0.034,
                                                    position_uncertainty=0.13,
                                                    gaussians_per_side=6,
                                                    horizontal=True,
                                                    additional_mean=np.array([5, 5, 0, 0, 0],
                                                                             dtype=np.float),
                                                    additional_cov=np.array([[50, 25, 0, 0, 0],
                                                                             [25, 50, 0, 0, 0],
                                                                             [0, 0, 1.5 * np.pi, 0, 0],
                                                                             [0, 0, 0, 7, 2.5],
                                                                             [0, 0, 0, 2.5, 7]]))

        # Survival Probability
        survival_probability: SurvivalProbability = self.create_survival_model(p_s=.9999999, fov=fov,
                                                                               horizontal=horizontal)

        # Transition Density

        scale: float = 1.1
        mat_a = np.eye(self.dimensionality)
        mat_a_t = np.zeros([self.dimensionality, self.dimensionality])
        mat_a_t[state_pos_columns[0], state_velocity_colum[0]] = \
            mat_a_t[state_pos_columns[1], state_velocity_colum[1]] = 0.0001
        model_a = LinearMotionModel(mat_a, mat_a_t)
        assert mat_a.shape == mat_a_t.shape
        assert mat_a.shape == (self.dimensionality, self.dimensionality)

        mat_q = np.diag([15 * 1.1, 47 * 1.1, 0.2, 1.5, np.pi / 32, 8.2, 8.2])
        mat_q_t = np.diag([0, 0, 0, 0, 0, 4, 4])
        model_q = LinearMotionModel(mat_q, mat_q_t)
        assert mat_q.shape == mat_q_t.shape
        assert mat_q.shape == (self.dimensionality, self.dimensionality)

        gaussian_transition_density = GaussianTransitionDensity(
            linear_motion_model=model_a, noise_covariance=model_q, rotate_noise_covariance=True,
            state_orientation_column=state_orientation_colum, survival_probability=survival_probability)

        predict = GaussianPredictionParametrization(birth=self.mbbm, transition=gaussian_transition_density)

        #
        # Update
        #
        transform = np.pad(np.eye(5), ((0, 0), (0, 2)))
        assert transform.shape == (self.dimensionality_z, self.dimensionality)

        update_cov = np.diag((45 * 1.5, 45 * 1.5, 0.4, 1.1, np.pi / 64))
        update_cov[1, 0] = update_cov[0, 1] = 2 * scale
        assert update_cov.shape == (self.dimensionality_z, self.dimensionality_z)

        update = UpdateModel(transform=transform, covariance=update_cov, p_d=0.2, p_c=5e-12)

        self.mbm = MultibernoulliMixture(predict=predict, update=update, max_number_of_hypothesis=1,
                                         detection_size_divided_by=self.detection_size_divisor)

    def process_item(self, item: Item) -> Item:

        t_start =time.time()

        # If the batch id changes, fire an empty item and reset mbm before processing the next element
        batch_id = item.get(['batch', 'id'])
        if batch_id != self.last_batch_id and self.mbm is not None:
            new_incoming, new_outgoing, incoming, outgoing, counted_objects_lifetime = self.mbm.reset()
            item_new = copy.deepcopy(item)
            item_new.setdefault('batch', dict())['id'] = self.last_batch_id

            item_new.setdefault('inference', dict())['counting'] = {
                'frame_incoming': new_incoming,
                'frame_outgoing': new_outgoing,
                'incoming': incoming,
                'outgoing': outgoing,
                'sum': incoming - outgoing,
                'counted_lifetime': counted_objects_lifetime
            }
            self.forward_item(item_new)
        self.last_batch_id = batch_id

        image = item.get('image')
        if self.mbm is None:
            if image is None:
                return item

            self.image_height, self.image_width = image.shape[:2]
            self.instanitate_mbm(image_width=self.image_width, image_height=self.image_height)

        current_time: float = item.get('timestamp', time.time())

        t_a = time.time()
        detections = self.extract_trackable_detections(item)
        confidences = detections[:, 5]

        if __debug__ and is_application_debug():
            assert ((confidences >= 0) * (confidences <= 1)).all()

        detections_all = self.extract_trackable_detections(item, all=True)
        t_b = time.time()

        t0 = time.time()
        estimates, tracks, new_in, new_out, total_in, total_out, display_image, counted_objects_lifetime = self.mbm.tic(
            dt=(current_time - self.t_last if self.t_last else 0),
            z=detections[:, :self.dimensionality_z],
            confidences=confidences,
            visualization_image=image.copy(),
            all_detections=detections_all)
        t1 = time.time()
        t_c = time.time()
        # print(t1 - t0)

        item.setdefault('inference', dict())['counting'] = {
            'frame_incoming': new_in,
            'frame_outgoing': new_out,
            'incoming': total_in,
            'outgoing': total_out,
            'sum': total_in - total_out,
            'counted_lifetime': counted_objects_lifetime
        }

        if total_in != self.last_total_in or  total_out != self.last_total_out:
            logger.info(f'in = {total_in}, out = {total_out}')
        self.last_total_in = total_in
        self.last_total_out = total_out

        item['display_image'] = display_image

        # item['display_image'] = self(item['image'], samples=detections_all, model=estimates, tracks=tracks)
        cv2.putText(item['display_image'], f'{total_in} {total_out}', (26, 25), cv2.FONT_HERSHEY_SIMPLEX, 0.6,
                    (250, 0, 250), 1, cv2.LINE_AA)

        if display_image is not None:
            item['image'] = item['display_image']

        # item['display_image'] = self(samples=detections[:,:2])

        self.t_last: float = current_time

        t_end =time.time()
        #print(t_end - t_start, t_b - t_a, t_c - t_b)

        return item

    @staticmethod
    def is_detection_trackable(detection, *, all: bool = False):
        return all or detection.get('confidence') >= 0.1

    def extract_trackable_detections(self, item: Item, all: bool = False) -> np.ndarray:
        detections: List[Dict] = list(
            filter(partial(self.is_detection_trackable, all=all), item.get(['inference', 'detection'], [])))
        new_detections: np.array = np.empty((len(detections), 6), dtype=np.float32)

        # Gather detection indices (index in the item's OG detection structure), position
        # size_scale = max(item['image'].shape[:2])

        for i, detection in enumerate(detections):
            x, y, width, height, orientation, confidence = detection.get('x'), detection.get('y'), \
                                                           detection.get('width'), detection.get('height'), \
                                                           detection.get('orientation'), detection.get('confidence')

            new_detections[i] = [x, y, width, height, orientation, confidence]

        np.multiply(new_detections[:, :4],
                    [self.image_width, self.image_height, self.image_width / self.detection_size_divisor,
                     self.image_height / self.detection_size_divisor],
                    out=new_detections[:, :4])

        # new_detections in 0, np.pi
        np.add(new_detections[:, 4], np.pi, out=new_detections[:, 4])  # geq 0
        np.mod(new_detections[:, 4], np.pi, out=new_detections[:, 4])  # in 0, pi
        # np.subtract(new_detections[:, 4], np.pi / 2, out=new_detections[:, 4])

        return new_detections

    @staticmethod
    def grid(image_width: int, image_height: int, resolution: float) -> np.ndarray:
        grid = np.mgrid[0: image_width: resolution, 0: image_height: resolution].T
        w, h = grid.shape[0], grid.shape[1]
        return grid.reshape([w * h, grid.shape[2]]), w, h

    def evaluate(self, xs, *, hypothesis_identifier: int = 0):
        return self.mbm(xs, hypothesis_identifier=hypothesis_identifier)

    def __call__(self, image, *, hypothesis_identifier: int = 0, plot: bool = False, samples=None, convert: bool = True,
                 model=None, tracks=None, evaluate_fully: bool = False, ):

        grid, w, h = self.grid(self.image_width, self.image_height, 1)

        if evaluate_fully:
            values = self.evaluate(grid, hypothesis_identifier=hypothesis_identifier)
            pmf = values.reshape((w, h))
        else:
            pmf = image.copy()
            convert = False
            # pmf = np.zeros((w, h))

        if convert:
            mn, mx = np.min(pmf), np.max(pmf)
            pmf -= mn
            if mx - mn:
                pmf /= (mx - mn)

            cmap = plt.get_cmap('jet')
            pmf = (np.delete(cmap(pmf), 3, 2) * 255).astype(np.uint8)

        if samples is not None:
            for marker in samples:
                if marker[2] < 0.1:
                    color = (255, 0, 0)
                elif marker[2] < .2:
                    color = (125, 0, 0)
                else:
                    color = (0, 0, 0)

                cv2.drawMarker(pmf, (marker[0], marker[1]), color, cv2.MARKER_CROSS, 30, 2)

        if model is not None:
            for tid, marker in model.items():

                color = get_track_color(tid)
                visualize_covariance: bool = False

                mean = marker[0]
                x, y = int(mean[0]), int(mean[1])

                if visualize_covariance:
                    covariance = marker[-1]

                    draw_covariance_ellises(pmf, means=mean[np.newaxis], covariance_matrices=covariance[np.newaxis],
                                            probabilities=[0.8, 0.9, 0.99], colors=[color], rs=None)

                # ptch.Ellipse((x, y), 2 * np.sqrt(varx) * v, 2 * np.sqrt(vary) * v)

                # cv2.ellipse(img=pmf, center=(int(x), int(y)), axes=(2 * varx * v, 2 * vary * v), angle=0, startAngle=0, endAngle=360, color=255, thickness=1, lineType=cv2.LINE_8, shift=0)

                # width = 2 * np.sqrt(varx) * v
                # height = 2 * np.sqrt(vary) * v

                # print(varx, vary, width, height, 'variance and width and height')

                # cv2.ellipse(pmf, (int(x), int(y)), (int(width), int(height)), 0, 0, 360, 255, 2, cv2.LINE_8, 0)

                x = min(max(0, x), self.image_width - 1)
                y = min(max(0, y), self.image_height - 1)
                cv2.drawMarker(pmf, (x, y), (100, 100, 100), cv2.MARKER_STAR, 15, 2)

        if tracks is not None:
            for i, (tid, track) in enumerate(tracks.items()):
                vs, ex, clr = track
                vs = vs.astype('int')
                for va, vb in zip(vs[1:], vs[:-1]):
                    cv2.line(pmf, tuple(va[:2]), tuple(vb[:2]), color=clr, thickness=1, lineType=cv2.LINE_AA)

                visualize_textual_information: bool = False
                if visualize_textual_information:
                    cv2.putText(pmf, f'{tid} {ex}', tuple(vs[-1][:2]), cv2.FONT_HERSHEY_SIMPLEX, 0.6, clr, 1,
                                cv2.LINE_AA)

        if plot:
            plt.figure(figsize=(self.image_width / 100, self.image_height / 100))
            plt.imshow(pmf)

            """

            if all_samples is not None:
                for sid, samples in all_samples.items():
                    plt.plot(*np.array(samples).T, color='black')

            if estimates is not None:
                for marker in estimates:
                    plt.plot(*marker, 'kx', markersize=10, color='red')

            if trajectories is not None:
                for tid, ts in trajectories.items():
                    plt.plot(*np.array(ts).T, color='red')
            """

            plt.show()

        return pmf


if __name__ == '__main__':
    tmbm = TrackingMbm(name=__name__)


    class F:
        def __init__(self):
            self.shape = (480, 640)


    tmbm.process_item(Item(image=F()))
